import enum

import karabo.native
import karathon
from karabo.common.scenemodel.api import (
    CheckBoxModel,
    ColorBoolModel,
    ComboBoxModel,
    DeviceSceneLinkModel,
    DetectorGraphModel,
    DisplayCommandModel,
    DisplayLabelModel,
    DisplayStateColorModel,
    DisplayTextLogModel,
    DoubleLineEditModel,
    EvaluatorModel,
    IntLineEditModel,
    LabelModel,
    LineEditModel,
    RectangleModel,
    SceneModel,
    SceneTargetWindow,
    write_scene,
)


# section: common setup


BASE_INC = 25
NARROW_INC = 20
PADDING = 5
RECONFIGURABLE = 4  # TODO: look up proper enum
NODE_TYPE_NODE = 1

_type_to_display_model = {
    "BOOL": CheckBoxModel
}
_type_to_line_editable = {
    "BOOL": (CheckBoxModel, {"klass": "EditableCheckBox"}),
    "DOUBLE": (DoubleLineEditModel, {}),
    "FLOAT": (DoubleLineEditModel, {}),
    "INT32": (IntLineEditModel, {}),
    "UINT32": (IntLineEditModel, {}),
    "INT64": (IntLineEditModel, {}),
    "UINT64": (IntLineEditModel, {}),
    "STRING": (LineEditModel, {"klass": "EditableLineEdit"}),
}


def safe_render(obj, x, y):
    if hasattr(obj, "render"):
        return obj.render(x, y)
    else:
        obj.x = x
        obj.y = y
        return [obj]


class Align(enum.Enum):
    CENTER = enum.auto()
    TOP = enum.auto()
    BOTTOM = enum.auto()
    LEFT = enum.auto()
    RIGHT = enum.auto()


# section: nice component decorators


def titled(title, width=8 * NARROW_INC):
    def actual_decorator(component_class):
        class new_class(component_class):
            def render(self, x, y, *args, **kwargs):
                return [
                    LabelModel(
                        frame_width=1,
                        text=title,
                        width=width,
                        height=NARROW_INC,
                        x=x,
                        y=y,
                    )
                ] + component_class.render(self, x, y + NARROW_INC, *args, **kwargs)

            @property
            def width(self):
                return max(component_class.width.fget(self), width)

            @property
            def height(self):
                return component_class.height.fget(self) + NARROW_INC

        return new_class

    return actual_decorator


def boxed(component_class):
    class new_class(component_class):
        def render(self, x, y, *args, **kwargs):
            return [
                RectangleModel(
                    x=x,
                    y=y,
                    width=component_class.width.fget(self) + 2 * PADDING,
                    height=component_class.height.fget(self) + 2 * PADDING,
                    stroke="#000000",
                )
            ] + component_class.render(self, x + PADDING, y + PADDING, *args, **kwargs)

        @property
        def width(self):
            return component_class.width.fget(self) + 2 * PADDING

        @property
        def height(self):
            return component_class.height.fget(self) + 2 * PADDING

    return new_class


# section: useful layout and utility classes


def DisplayRoundedFloat(*args, decimals=2, **kwargs):
    # note: naive subclass breaks as registry looks for writer based on exact class
    return EvaluatorModel(*args, expression=f"f'{{x:{decimals}}}'", **kwargs)


class Space:
    def __init__(self, width, height):
        self.width = width
        self.height = height

    def render(self, x, y):
        return []


def dummy_wrap(model_class):
    class Wrapper:
        def __init__(self, *args, **kwargs):
            self.thing = model_class(*args, **kwargs)

        def render(self, x, y):
            self.thing.x = x
            self.thing.y = y
            return [self.thing]

        @property
        def width(self):
            return self.thing.width

        @property
        def height(self):
            return self.thing.height

    return Wrapper


class HorizontalLayout:
    def __init__(self, *arg_children, children=None, padding=PADDING):
        self.children = list(arg_children)
        if children is not None:
            self.children.extend(children)
        self.padding = padding

    def render(self, x, y, align=Align.TOP):
        if align is not Align.TOP:
            height = self.height
        res = []
        for child in self.children:
            if align is Align.TOP:
                y_ = y
            elif align is Align.CENTER:
                y_ = y + (height - child.height) / 2
            elif align is Align.BOTTOM:
                y_ = y + (height - child.height)
            else:
                raise ValueError(f"Invalid align {align} for HorizontalLayout")
            res.extend(safe_render(child, x, y_))
            x += child.width + self.padding
        return res

    @property
    def width(self):
        if not self.children:
            return 0
        return self.padding * (len(self.children) - 1) + sum(
            c.width for c in self.children
        )

    @property
    def height(self):
        if not self.children:
            return 0
        return max(c.height for c in self.children)


class VerticalLayout:
    def __init__(self, *arg_children, children=None, padding=PADDING):
        self.children = list(arg_children)
        if children is not None:
            self.children.extend(children)
        self.padding = padding

    def render(self, x, y):
        res = []
        for child in self.children:
            res.extend(safe_render(child, x, y))
            y += child.height + self.padding
        return res

    @property
    def width(self):
        if not self.children:
            return 0
        return max(c.width for c in self.children)

    @property
    def height(self):
        if not self.children:
            return 0
        return self.padding * (len(self.children) - 1) + sum(
            c.height for c in self.children
        )


class MaybeEditableRow(HorizontalLayout):
    def __init__(
        self,
        device_id,
        schema_hash,
        key_path,
        label_width=7 * NARROW_INC,
        display_width=5 * NARROW_INC,
        edit_width=5 * NARROW_INC,
        height=NARROW_INC,
    ):
        super().__init__(padding=0)
        key_attr = schema_hash.getAttributes(key_path)
        label_text = (
            key_attr["displayedName"]
            if "displayedName" in key_attr
            else key_path.split(".")[-1]
        )
        if "valueType" not in key_attr:
            print(f"Key {key_path} on {device_id} had no valueType")
            return
        value_type = key_attr["valueType"]
        self.children.extend(
            [
                LabelModel(
                    text=label_text,
                    width=label_width,
                    height=height,
                ),
                _type_to_display_model.get(value_type, DisplayLabelModel)(
                    keys=[f"{device_id}.{key_path}"],
                    width=display_width,
                    height=height,
                ),
            ]
        )
        if key_attr["accessMode"] == RECONFIGURABLE:
            if "options" in key_attr:
                self.children.append(
                    ComboBoxModel(
                        keys=[f"{device_id}.{key_path}"],
                        width=edit_width,
                        height=height,
                        klass="EditableComboBox",
                    )
                )
            elif value_type in _type_to_line_editable:
                line_editable_class, extra_args = _type_to_line_editable[value_type]
                self.children.append(
                    line_editable_class(
                        keys=[f"{device_id}.{key_path}"],
                        width=edit_width,
                        height=height,
                        **extra_args,
                    )
                )
            else:
                self.children.append(
                    LabelModel(
                        text=f"Not implemented: editing {value_type} ({key_path})",
                        width=edit_width,
                        height=height,
                    )
                )


# section: specific handcrafted components for device classes


@titled("Found constants", width=6 * NARROW_INC)
@boxed
class FoundConstantsColumn(VerticalLayout):
    def __init__(self, device_id, schema_hash, prefix="foundConstants"):
        super().__init__(padding=0)
        self.children.extend(
            [
                HorizontalLayout(
                    LabelModel(
                        text=constant_name,
                        width=6 * NARROW_INC,
                        height=NARROW_INC,
                    ),
                    ColorBoolModel(
                        width=NARROW_INC,
                        height=NARROW_INC,
                        keys=[f"{device_id}.{prefix}.{constant_name}.found"],
                    ),
                    DisplayLabelModel(
                        keys=[f"{device_id}.{prefix}.{constant_name}.validFrom"],
                        width=8 * BASE_INC,
                        height=BASE_INC,
                    ),
                    padding=0,
                )
                for constant_name in schema_hash.get(prefix).getKeys()
            ]
        )


class ConstantLoadedAmpeln(HorizontalLayout):
    def __init__(self, device_id, schema_hash, prefix="foundConstants"):
        super().__init__(padding=0)
        self.children.extend(
            [
                ColorBoolModel(
                    keys=[f"{device_id}.{prefix}.{key}.found"],
                    height=BASE_INC,
                    width=BASE_INC,
                )
                for key in schema_hash.get(prefix).getKeys()
            ]
        )


@titled("Manager status", width=6 * NARROW_INC)
@boxed
class ManagerDeviceStatus(VerticalLayout):
    def __init__(self, device_id):
        super().__init__(padding=0)
        name = DisplayLabelModel(
            keys=[f"{device_id}.deviceId"],
            width=14 * BASE_INC,
            height=BASE_INC,
        )
        state = DisplayStateColorModel(
            show_string=True,
            keys=[f"{device_id}.state"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        restart_button = DisplayCommandModel(
            keys=[f"{device_id}.restartServers"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        instantiate_button = DisplayCommandModel(
            keys=[f"{device_id}.startInstantiate"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        apply_button = DisplayCommandModel(
            keys=[f"{device_id}.applyManagedValues"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        status_log = DisplayTextLogModel(
            keys=[f"{device_id}.status"],
            width=14 * BASE_INC,
            height=14 * BASE_INC,
        )
        self.children.extend(
            [
                name,
                HorizontalLayout(
                    state,
                    restart_button,
                    padding=0,
                ),
                HorizontalLayout(
                    instantiate_button,
                    apply_button,
                    padding=0,
                ),
                DeviceSceneLinkModel(
                    text="All managed properties",
                    keys=[f"{device_id}.availableScenes"],
                    target="browse_schema",
                    target_window=SceneTargetWindow.Dialog,
                    width=7 * BASE_INC,
                    height=BASE_INC,
                ),
                status_log,
            ]
        )


@titled("Device status", width=6 * NARROW_INC)
@boxed
class CorrectionDeviceStatus(VerticalLayout):
    def __init__(self, device_id):
        super().__init__(padding=0)
        name = DisplayLabelModel(
            keys=[f"{device_id}.deviceId"],
            width=14 * BASE_INC,
            height=BASE_INC,
        )
        state = DisplayStateColorModel(
            show_string=True,
            keys=[f"{device_id}.state"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        rate = DisplayRoundedFloat(
            keys=[f"{device_id}.performance.rate"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        processing_time = DisplayRoundedFloat(
            keys=[f"{device_id}.performance.processingTime"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        tid = DisplayLabelModel(
            keys=[f"{device_id}.trainId"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        status_log = DisplayTextLogModel(
            keys=[f"{device_id}.status"],
            width=14 * BASE_INC,
            height=14 * BASE_INC,
        )
        self.children.extend(
            [
                name,
                HorizontalLayout(
                    state,
                    tid,
                    padding=0,
                ),
                HorizontalLayout(
                    rate,
                    processing_time,
                    padding=0,
                ),
                status_log,
            ]
        )


class CompactCorrectionDeviceOverview(HorizontalLayout):
    def __init__(self, device_id, schema_hash):
        super().__init__(padding=0)
        self.children.extend(
            [
                DeviceSceneLinkModel(
                    text=device_id.split("/")[-1],
                    keys=[f"{device_id}.availableScenes"],
                    target="overview",
                    target_window=SceneTargetWindow.Dialog,
                    width=5 * BASE_INC,
                    height=BASE_INC,
                ),
                DisplayStateColorModel(
                    show_string=True,
                    keys=[f"{device_id}.state"],
                    width=6 * BASE_INC,
                    height=BASE_INC,
                ),
                DisplayRoundedFloat(
                    keys=[f"{device_id}.performance.rate"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
                DisplayLabelModel(
                    keys=[f"{device_id}.trainId"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
                ConstantLoadedAmpeln(device_id, schema_hash),
            ]
        )


@titled("Other devices managed")
@boxed
class CompactDeviceLinkList(VerticalLayout):
    def __init__(self, device_ids):
        super().__init__()
        self.children.extend(
            [
                HorizontalLayout(
                    DeviceSceneLinkModel(
                        text=device_id.split("/")[-1],
                        keys=[f"{device_id}.availableScenes"],
                        width=7 * BASE_INC,
                        height=BASE_INC,
                    ),
                    DisplayStateColorModel(
                        show_string=True,
                        keys=[f"{device_id}.state"],
                        width=7 * BASE_INC,
                        height=BASE_INC,
                    ),
                    padding=0,
                )
                for device_id in device_ids
            ]
        )


@titled("Assembler status", width=8 * NARROW_INC)
@boxed
class AssemblerDeviceStatus(VerticalLayout):
    def __init__(self, device_id):
        super().__init__(padding=0)
        name = DisplayLabelModel(
            keys=[f"{device_id}.deviceId"],
            width=14 * BASE_INC,
            height=BASE_INC,
        )
        state = DisplayStateColorModel(
            show_string=True,
            keys=[f"{device_id}.state"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        train_id = DisplayLabelModel(
            keys=[f"{device_id}.trainId"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        self.children.extend(
            [
                name,
                HorizontalLayout(
                    state,
                    train_id,
                    padding=0,
                ),
                LabelModel(
                    text="Image downsampling",
                    width=14 * BASE_INC,
                    height=BASE_INC,
                ),
                HorizontalLayout(
                    LabelModel(
                        text="Factor",
                        width=7 * BASE_INC,
                        height=BASE_INC,
                    ),
                    ComboBoxModel(
                        keys=[f"{device_id}.downsamplingFactor"],
                        width=7 * BASE_INC,
                        height=BASE_INC,
                        klass="EditableComboBox",
                    ),
                    padding=0,
                ),
                HorizontalLayout(
                    LabelModel(
                        text="Function",
                        width=7 * BASE_INC,
                        height=BASE_INC,
                    ),
                    ComboBoxModel(
                        keys=[f"{device_id}.downsamplingFunction"],
                        width=7 * BASE_INC,
                        height=BASE_INC,
                        klass="EditableComboBox",
                    ),
                    padding=0,
                ),
                DeviceSceneLinkModel(
                    text="I'm actually a TrainMatcher",
                    keys=[f"{device_id}.availableScenes"],
                    target="trainMatcherScene",
                    target_window=SceneTargetWindow.Dialog,
                    width=14 * BASE_INC,
                    height=BASE_INC,
                ),
            ]
        )


# section: generating actual scenes


def schema_to_hash(schema):
    if isinstance(schema, (karathon.Hash, karabo.native.Hash)):
        return schema
    elif isinstance(schema, karathon.Schema):
        return schema.getParameterHash()
    else:
        return schema.hash


def scene_generator(fun):
    # TODO: pretty decorator
    def aux(*args, **kwargs):
        content = fun(*args, **kwargs)

        scene = SceneModel(
            children=content.render(PADDING, PADDING),
            width=content.width + 2 * PADDING,
            height=content.height + 2 * PADDING,
        )
        return write_scene(scene)

    return aux


@scene_generator
def correction_device_overview_scene(device_id, schema):
    schema_hash = schema_to_hash(schema)

    return HorizontalLayout(
        CorrectionDeviceStatus(device_id),
        VerticalLayout(
            recursive_maybe_editable(
                device_id,
                schema_hash,
                "constantParameters",
            ),
            DisplayCommandModel(
                keys=[f"{device_id}.loadMostRecentConstants"],
                width=10 * BASE_INC,
                height=BASE_INC,
            ),
        ),
        FoundConstantsColumn(device_id, schema_hash),
        recursive_maybe_editable(
            device_id,
            schema_hash,
            "corrections",
            max_depth=2,
        ),
    )


@scene_generator
def manager_device_overview_scene(
    manager_device_id,
    manager_device_schema,
    correction_device_schema,
    correction_device_ids,
    domain_device_ids,
):
    mds_hash = schema_to_hash(manager_device_schema)
    cds_hash = schema_to_hash(correction_device_schema)

    return VerticalLayout(
        HorizontalLayout(
            ManagerDeviceStatus(manager_device_id),
            VerticalLayout(
                recursive_maybe_editable(
                    manager_device_id,
                    mds_hash,
                    "managed.constantParameters",
                ),
                DisplayCommandModel(
                    keys=[f"{manager_device_id}.managed.loadMostRecentConstants"],
                    width=10 * BASE_INC,
                    height=BASE_INC,
                ),
                recursive_maybe_editable(
                    manager_device_id,
                    mds_hash,
                    "managed.preview",
                    max_depth=2,
                ),
            ),
            recursive_maybe_editable(
                manager_device_id,
                mds_hash,
                "managed.corrections",
                max_depth=2,
            ),
        ),
        HorizontalLayout(
            titled("Correction devices", width=8 * NARROW_INC)(boxed(VerticalLayout))(
                children=[
                    CompactCorrectionDeviceOverview(device_id, cds_hash)
                    for device_id in sorted(correction_device_ids)
                ],
                padding=0,
            ),
            CompactDeviceLinkList(
                sorted(
                    set(domain_device_ids)
                    - set(correction_device_ids)
                    - {manager_device_id}
                )
            ),
        ),
    )

@scene_generator
def simple_assembler_overview(device_id, geometry_device_id):
    return VerticalLayout(
        HorizontalLayout(
            AssemblerDeviceStatus(device_id),
            titled("My geometry device")(boxed(VerticalLayout))(
                DeviceSceneLinkModel(
                    text=f"Geometry device: {geometry_device_id}",
                    keys=[f"{geometry_device_id}.availableScenes"],
                    target="overview",
                    target_window=SceneTargetWindow.Dialog,
                    frame_width=1,
                    width=14 * BASE_INC,
                    height=BASE_INC,
                ),
                DisplayCommandModel(
                    keys=[f"{geometry_device_id}.pleaseSendYourGeometry"],
                    width=14 * BASE_INC,
                    height=BASE_INC,
                ),
                padding=0,
            ),
        ),
        titled("Preview image")(boxed(dummy_wrap(DetectorGraphModel)))(
            keys=[f"{device_id}.preview.output.schema.image"],
            colormap="viridis",
            width=30 * BASE_INC,
            height=30 * BASE_INC,
            x=PADDING,
            y=PADDING,
        ),
    )


# section: here be monsters


def recursive_maybe_editable(
    device_id, schema_hash, prefix, depth=1, max_depth=3, title=None
):
    schema_hash = schema_to_hash(schema_hash)
    # note: not just using sets because that loses ordering
    node_keys = []
    value_keys = []
    slot_keys = []
    attr = schema_hash.getAttributes(prefix)

    if title is None:
        if "displayedName" in attr:
            title = attr.get("displayedName")
        else:
            title = prefix.split(".")[-1]

    for key in schema_hash.get(prefix).getKeys():
        attrs = schema_hash.getAttributes(f"{prefix}.{key}")
        if attrs.get("nodeType") == NODE_TYPE_NODE:
            if "classId" in attrs and attrs.get("classId") == "Slot":
                slot_keys.append(key)
            else:
                node_keys.append(key)
        else:
            value_keys.append(key)
    res = titled(title)(boxed(VerticalLayout))(
        children=[
            MaybeEditableRow(device_id, schema_hash, f"{prefix}.{key}")
            for key in value_keys
        ]
        + [
            DisplayCommandModel(
                keys=[f"{device_id}.{prefix}.{key}"],
                width=10 * BASE_INC,
                height=BASE_INC,
            )
            for key in slot_keys
        ],
        padding=0,
    )
    if depth < max_depth:
        res.children.append(
            VerticalLayout(
                children=[
                    recursive_maybe_editable(
                        device_id,
                        schema_hash,
                        f"{prefix}.{key}",
                        depth=depth + 1,
                        max_depth=max_depth,
                    )
                    for key in node_keys
                ]
            )
        )
    else:
        res.children.extend(
            [
                VerticalLayout(
                    DeviceSceneLinkModel(
                        text=key,
                        keys=[f"{device_id}.availableScenes"],
                        target=f"browse_schema:{prefix}.{key}",
                        target_window=SceneTargetWindow.Dialog,
                        width=5 * BASE_INC,
                        height=BASE_INC,
                    ),
                )
                for key in node_keys
            ]
        )
    return res


@scene_generator
def recursive_subschema_scene(
    device_id,
    device_schema,
    prefix="managed",
):
    mds_hash = schema_to_hash(device_schema)
    return recursive_maybe_editable(device_id, mds_hash, prefix)
